import { NextApiRequest, NextApiResponse } from "next"
import logging, { Logging } from "../../lib/backend/logging"
import getPlayerByNameDB from "../../lib/backend/components/getPlayerByNameDB"
import checkPasswordIsValid from "../../lib/backend/components/checkPasswordIsValid"
import createTokenDB from "../../lib/backend/components/createTokenDB"
import sendResponse from "../../lib/backend/components/sendResponse"
import sendError from "../../lib/backend/components/sendError"
import { deleteCookie, setCookie } from "cookies-next"
import { Player, Token } from "@prisma/client"
import prisma from "../../lib/prisma"
import jwt from "jsonwebtoken"

interface Data {
  loggedIn: boolean
}

export default async function login(
  req: NextApiRequest,
  res: NextApiResponse<any>
) {
  const { username, password } = JSON.parse(req.body)
  const payload = {
    req,
    res,
    username,
    password,
    newTokenType: "REFRESH" as Token["type"],
  }
  return preCheck(payload)
    .then(getPlayerByNameDB)
    .then(checkPasswordIsValid)
    .then(createTokenDB)
    .then(loginResponse<Data>)
    .then(sendResponse<Data>)
    .catch((err) => sendError(req, res, err))
}

async function preCheck<T>(
  payload: T & {
    req: NextApiRequest
    res: NextApiResponse<T>
  }
) {
  const { req, res } = payload
  const oldRefreshToken = req.cookies.token
  // Check for old cookie, if unused invalidate it
  const tokenData =
    oldRefreshToken &&
    jwt.verify(oldRefreshToken, process.env.ACCESS_TOKEN_SECRET as string)

  if (!tokenData || typeof tokenData === "string")
    deleteCookie("token", { req, res })
  else {
    const oldDBToken =
      oldRefreshToken &&
      (await prisma.token.findUnique({
        where: {
          id: tokenData.id,
        },
      }))
    if (!oldDBToken) return
    if (!oldDBToken.used)
      logging("Old token was used: " + oldDBToken.id, ["debug"], req)
    else {
      await prisma.token.update({
        where: {
          id: tokenData.id,
        },
        data: {
          used: true,
        },
      })
    }
    await logging("Old token has been invalidated.", ["debug"], req)
  }
  return { ...payload, noCookiePresent: true }
}

async function loginResponse<T>(payload: {
  player: Player
  passwordIsValid: boolean
  newToken: string
  newTokenDB: Token
  req: NextApiRequest
  res: NextApiResponse<T>
}) {
  const { player, newToken, newTokenDB, req, res } = payload

  // Set login cookie
  setCookie("token", newToken, {
    req,
    res,
    maxAge: 172800000,
    httpOnly: true,
    sameSite: true,
    secure: true,
  })

  // Successfull response
  return {
    req,
    res,
    result: {
      message:
        "User " +
        player.id +
        " logged in and generated Refresh-Token: " +
        newTokenDB.id,
      body: { loggedIn: true },
      type: ["debug", "info.cyan"] as Logging[],
    },
  }
}
